# Aggregation Window
SCALING_FACTOR =0.1
BOX_WIDHT = 300
BOX_HEIGHT = 200
BOX_MARGIN = 60
BOX_BOTTOM_DISTANCE = 40
ARROW_HEIGHT = 20
ARROW_WIDTH = ARROW_HEIGHT/2
AGGREGATION = 1
INHERITANCE = 2

DESCRIPTION = "description"
OPTIONAL = "optional"

VALUE = "Value"
FORMAT = "Format"
RANGE = "Range"
LIST = "List"
XS_STRING = "xs:string"
XS_DOUBLE = "xs:double"
XS_BOOL = "xs:boolean"
XS_INT = "xs:int"

RANGE_STRINGS = ["Range", "range", "RANGE"]

VALUE_TYPE_LOOKUP = {VALUE: "Value", FORMAT: "Format", RANGE: "Range", LIST: "List", }
PREDEFINED_PROPERTY_WINDOW_NAME = "Predefined Properties"

DATATYPE_NUMBER = "xs:double"
ICON_PATH = "icons/icon.ico"
LINK_ICON_PATH = "icons/link.png"

DATA_POS = 1
FILEPATH_JS = "js_templates"

IGNORE_PSET = "IFC"
INHERITED_TEXT = "Predefined Pset"

# xml export
PREDEFINED_PSETS = "PredefinedPropertySets"
PREDEFINED_PSET = "PredefinedPropertySet"
PROPERTY_SET = "PropertySet"
PROPERTY_SETS = "PropertySets"
IDENT_ATTRIBUTE = "ident_attribute"
ATTRIBUTE = "Attribute"
ATTRIBUTES = "Attributes"
OBJECT = "Object"
OBJECTS = "Objects"
AGGREGATE = "Aggregate"
AGGREGATIONS = "Aggregations"
NAME = "name"
IDENTIFIER = "identifer"
PARENT = "parent"
NONE = "None"
DATA_TYPE = "data_type"
VALUE_TYPE = "value_type"
IS_IDENTIFIER = "is_identifier"
CHILD_INHERITS_VALUE = "child_inherits_value"
IS_CONCEPT = "is_concept"
SCRIPT = "Script"
SCRIPTS = "Scripts"
PROJECT = "Project"
VERSION = "version"
AGGREGATES_TO = "aggregates_to"
AUTHOR = "author"
X_POS = "x_pos"
Y_POS = "y_pos"
IS_ROOT = "root"
CONNECTION = "connection"
NODE = "Node"
NODES = "Nodes"
IFC_MAPPINGS = "IfcMappings"
IFC_MAPPING = "IfcMapping"
CUSTOM_ATTRIBUTES = "CustomAttributes"
CUSTOM_ATTRIBUTE = "CustomAttribute"
ABBREVIATION = "abbreviation"
#Mapping
SHARED_PARAMETERS = "SharedParameters"
REVIT_MAPPING = "revit_mapping"


import os

from lxml import etree, builder
from openpyxl import load_workbook

from .. import classes


def create_mapping(excel_path: str, folder_path: str, project:classes.Project) -> None:
    def create_xml(name: str, object: classes.Object):
        xsd = "http://www.w3.org/2001/XMLSchema"
        xsi = "http://www.w3.org/2001/XMLSchema-instance"

        def create_key_rule() -> None:
            xml_key_rule = etree.SubElement(xml_manipulation_rule, "KeyRule", )
            xml_key_rule.set(f"{{{xsi}}}type", "CompareRule")
            xml_key = etree.SubElement(xml_key_rule, "Key", )
            xml_key.text = "Querprofile-Horizontname"
            xml_value = etree.SubElement(xml_key_rule, "Value", )
            xml_value.text = name
            xml_value.set(f"{{{xsi}}}type", "xsd:string")
            xml_value = etree.SubElement(xml_key_rule, "Compare", )
            xml_value.text = "Equal"

        def create_manipulations() -> None:
            xml_manipulations = etree.SubElement(xml_manipulation_rule, "Manipulations")
            for property_set in object.property_sets:
                for attribut in property_set.attributes:
                    xml_manipulation_base = etree.SubElement(xml_manipulations, "ManipulationBase")
                    xml_manipulation_base.set(f"{{{xsi}}}type", "AddManipulation")
                    xml_key = etree.SubElement(xml_manipulation_base, "Key", )
                    xml_key.text = attribut.name
                    xml_value = etree.SubElement(xml_manipulation_base, "Value", )
                    if attribut == object.ident_attrib:
                        xml_value.text = attribut.value[0]
                    else:
                        xml_value.text = ""
                    xml_value.set(f"{{{xsi}}}type", attribut.data_type.replace("xs:", "xsd:"))

        E = builder.ElementMaker(nsmap={"xsd": xsd, "xsi": xsi})
        xml_role = E.Role()
        xml_name = etree.SubElement(xml_role, "Name")
        xml_name.text = name
        xml_rules = etree.SubElement(xml_role, "Rules")
        xml_manipulation_rule = etree.SubElement(xml_rules, "ManipulationRule")
        xml_description = etree.SubElement(xml_manipulation_rule, "Description")
        xml_description.text = "Autogenerated"
        create_key_rule()
        create_manipulations()

        tree = etree.ElementTree(xml_role)
        tree.write(os.path.join(folder_path, f"{name.replace('/', '_')}.amrx"), pretty_print=True, encoding="utf-8",
                   xml_declaration=True)

    wb = load_workbook(excel_path)
    sheet = wb.active
    if not os.path.isdir(folder_path):
        os.mkdir(folder_path)

    important_rows = [row for i, row in enumerate(sheet.rows) if row[2].value is not None and i != 0]
    object_dict = {obj.ident_attrib.value[0]: obj for obj in project.objects if not obj.is_concept}

    for row in important_rows:
        values = list(map(lambda x: x.value, row))
        if len(values)!= 4:
            raise ValueError("Spaltenkonfiguration nicht korrekt!")
        hz_nummer, hz_name, bauteilName,bauteilklass =values
        create_xml(hz_name, object_dict[bauteilklass])



from __future__ import annotations

from typing import TYPE_CHECKING, IO, Iterator

from .. import classes, constants

if TYPE_CHECKING:
    pass


def export_ifc_template(path: str, pset_dict: dict[str, (list[classes.Attribute], set[str])]) -> None:
    def transform_datatype(data_type: str) -> str:
        if data_type == constants.XS_INT:
            return "Integer"
        if data_type == constants.XS_STRING:
            return "Label"
        if data_type == constants.XS_DOUBLE:
            return "Real"
        if data_type == constants.XS_BOOL:
            return "Boolean"
        return "ERROR"

    with open(path, "w") as file:
        property_set: classes.PropertySet
        for pset_name, (attrib_list, ifc_mapping) in sorted(pset_dict.items()):
            file.write(f"PropertySet:   {pset_name} I  {','.join(ifc_mapping)} \n")
            for attribute in attrib_list:
                revit_datatype = transform_datatype(attribute.data_type)
                file.write(f"   {attribute.name}    {revit_datatype}\n")
            file.write("\n")


class IterItem(type):
    _registry = set()

    def __iter__(cls) -> Iterator[SP_Item]:
        return iter(cls._registry)

    def add_item(cls, item):
        cls._registry.add(item)

    def __len__(cls):
        return len(cls._registry)

    # def __new__(meta,name,bases,attrs):
    #     attrs['_cars'] = weaker.WeakSet()
    #     return type.__new__(meta, name, bases, attrs)


class SP_Item(metaclass=IterItem):

    def __init__(self, property_set_name, attribute, pset_number):
        self.__class__.add_item(self)
        self.property_set_name = property_set_name
        self.attribute = attribute
        self.pset_number = pset_number

    def print(self, file: IO):
        file.write(
            f"PARAM	{self.attribute.uuid}	{self.attribute.name}	{self.datatype()}		{self.pset_number}	1		1\n")

    def datatype(self) -> str:
        data_type = self.attribute.data_type
        if data_type == constants.XS_INT:
            return "INTEGER"
        if data_type == constants.XS_STRING:
            return "TEXT"
        if data_type == constants.XS_DOUBLE:
            return "NUMBER"
        if data_type == constants.XS_BOOL:
            return "YESNO"
        return "ERROR"


def export_shared_parameters(path: str, pset_dict: dict[str, (list[classes.Attribute], set[str])]) -> None:
    with open(path, "w") as file:
        file.write("# This is a Revit shared parameter file.\n"
                   "# Do not edit manually.\n"
                   "*META	VERSION	MINVERSION\n"
                   "META	2	1\n"
                   "*GROUP	ID	NAME\n")

        for i, pset_name in enumerate(sorted(pset_dict.keys())):
            file.write(f"GROUP	{i + 1}	{pset_name}\n")

        file.write(
            "*PARAM	GUID	NAME	DATATYPE	DATACATEGORY	GROUP	VISIBLE	DESCRIPTION	USERMODIFIABLE\n")

        property_set: classes.PropertySet
        for i, (pset_name, (attrib_list, ifc_mapping)) in enumerate(sorted(pset_dict.items())):
            for attrib in attrib_list:
                t = SP_Item(pset_name, attrib, i)

        for item in sorted(SP_Item, key=lambda x: x.attribute.name):
            item.print(file)
            pass



from __future__ import annotations

import codecs
import csv
import datetime
import json
import os
import uuid
import xml.etree.ElementTree as ET

import jinja2
from anytree import AnyNode
from lxml import etree

from .. import classes, constants, Template
from ..Template import HOME_DIR,BOOKMARK_TEMPLATE
output_date_time = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
output_date = datetime.datetime.now().strftime("%Y-%m-%d")


def handle_header(project: classes.Project, export_format: str) -> etree._Element:
    ET.register_namespace("xsi", "http://www.w3.org/2001/XMLSchema-instance")
    xml_header = etree.Element(f'{{http://www.w3.org/2001/XMLSchema-instance}}{export_format}')
    xml_header.set("user", str(project.author))
    xml_header.set("date", str(output_date_time))
    xml_header.set("version", "3.0.1")  # TODO: Desite version hinzufügen
    return xml_header


##TODO add xs:bool

def export_modelcheck(project, path: str, project_tree=None) -> None:
    def add_js_rule(parent: etree._Element, file: codecs.StreamReaderWriter) -> str | None:
        name = os.path.basename(file.name)
        if not name.endswith(".js"):
            return None
        else:
            rule_script = etree.SubElement(parent, "ruleScript")

            name = name.split("_")[1:]
            name = "_".join(name)
            rule_script.set("name", name[:-3])
            rule_script.set("active", "true")
            rule_script.set("resume", "false")

            code = etree.SubElement(rule_script, "code")
            file = file.read()

            code.text = file
            code.text = etree.CDATA(code.text)

        return file

    def handle_element_section(xml_qa_export: etree._Element) -> etree._Element:
        xml_element_section = etree.SubElement(xml_qa_export, "elementSection")
        return xml_element_section

    def handle_container(xml_element_section: etree._Element, text) -> etree._Element:
        container = etree.SubElement(xml_element_section, "container")
        container.set("ID", str(uuid.uuid4()))
        container.set("name", text)
        return container

    def handle_checkrun(xml_container: etree._Element, name: str, author: str = "DesiteRuleCreator") -> etree._Element:
        checkrun = etree.SubElement(xml_container, "checkrun")
        _uuid = str(uuid.uuid4())
        checkrun.set("ID", _uuid)
        checkrun.set("name", name)
        checkrun.set("active", "true")
        checkrun.set("user", str(author))
        checkrun.set("date", str(output_date_time))
        checkrun.set("state", "0")
        checkrun.set("objectsOnly", "1")
        checkrun.set("partsOfComposites", "0")
        checkrun.set("createFailed", "true")
        checkrun.set("createWarnings", "true")
        checkrun.set("createIgnored", "false")
        checkrun.set("createPassed", "true")
        checkrun.set("createUndefined", "false")
        return checkrun

    def init_xml(project) -> (etree._Element, etree._Element):
        xml_qa_export = handle_header(project, "qaExport")
        xml_element_section = handle_element_section(xml_qa_export)
        text = f"{project.name} : {project.version}"
        xml_container = handle_container(xml_element_section, text)
        return xml_container, xml_qa_export

    def handle_rule(xml_checkrun: etree._Element, rule_type: str) -> etree._Element:
        rule = etree.SubElement(xml_checkrun, "rule")
        rule.set("type", rule_type)
        if rule_type == "UniquePattern":
            etree.SubElement(rule, "patternList")
            code = etree.SubElement(xml_checkrun, "code")
            code.text = ""

        return rule

    def handle_attribute_rule_list(xml_rule: etree._Element) -> etree._Element:
        attribute_rule_list = etree.SubElement(xml_rule, "attributeRuleList")
        return attribute_rule_list

    def handle_template() -> jinja2.Template:
        path = Template.HOME_DIR
        file_loader = jinja2.FileSystemLoader(path)
        env = jinja2.Environment(loader=file_loader)
        env.trim_blocks = True
        env.lstrip_blocks = True
        template = env.get_template("template.txt")

        return template

    def define_xml_elements(xml_container: etree._Element, name: str) -> (etree._Element, etree._Element):
        xml_checkrun = handle_checkrun(xml_container, name=name, author=project.author)
        xml_rule = handle_rule(xml_checkrun, "Attributes")
        xml_attribute_rule_list = handle_attribute_rule_list(xml_rule)
        handle_rule(xml_checkrun, "UniquePattern")

        return xml_checkrun, xml_attribute_rule_list

    def handle_js_rules(xml_attribute_rule_list: etree._Element, starts_with: str) -> None:
        folder = os.path.join(Template.HOME_DIR, constants.FILEPATH_JS)

        for fn in os.listdir(folder):
            if fn.startswith(starts_with):
                file = codecs.open(f"{folder}/{fn}", encoding="utf-8")
                add_js_rule(xml_attribute_rule_list, file)

    def handle_rule_script(xml_attribute_rule_list: etree._Element, name: str) -> etree._Element:
        rule_script = etree.SubElement(xml_attribute_rule_list, "ruleScript")
        rule_script.set("name", name)
        rule_script.set("active", "true")
        rule_script.set("resume", "true")
        return rule_script

    def handle_code(xml_rule_script: etree._Element) -> etree._Element:
        code = etree.SubElement(xml_rule_script, "code")
        return code

    def handle_object_rules(base_xml_container: etree._Element, template: jinja2.Template) -> dict[
        etree._Element, classes.Object]:

        def handle_tree_structure(parent_xml_container, parent_node: AnyNode) -> None:

            def create_container(xml_container, node: AnyNode):
                new_xml_container = handle_container(xml_container, node.obj.name)
                create_object(new_xml_container, node)
                for child_node in sorted(node.children, key=lambda x: x.id):
                    handle_tree_structure(new_xml_container, child_node)

            def create_object(xml_container, node: AnyNode):
                obj: classes.Object = node.obj
                xml_checkrun = handle_checkrun(xml_container, obj.name, project.author)
                xml_rule = handle_rule(xml_checkrun, "Attributes")
                xml_attribute_rule_list = handle_attribute_rule_list(xml_rule)
                xml_rule_script = handle_rule_script(xml_attribute_rule_list, name=obj.name)
                xml_code = handle_code(xml_rule_script)

                property_sets = [pset for pset in obj.property_sets if len(pset.attributes) > 0]
                ident_name = obj.ident_attrib.name
                ident_property_set = obj.ident_attrib.property_set.name
                if ident_property_set == constants.IGNORE_PSET:
                    ident_property_set = ""
                else:
                    ident_property_set = f"{ident_property_set}:"

                cdata_code = template.render(psets=property_sets, object=obj, ident=ident_name,
                                             ident_pset=ident_property_set, constants=constants)
                xml_code.text = cdata_code
                handle_rule(xml_checkrun, "UniquePattern")

                for script in obj.scripts:
                    xml_rule_script = handle_rule_script(xml_attribute_rule_list, name=script.name)
                    xml_code = handle_code(xml_rule_script)
                    xml_code.text = script.code

                xml_object_dict[xml_checkrun] = obj

            if parent_node.children:
                create_container(parent_xml_container, parent_node)
            else:
                create_object(parent_xml_container, parent_node)

        xml_object_dict: dict[etree._Element, classes.Object] = dict()
        root_nodes = project_tree.children

        for root_node in sorted(root_nodes, key=lambda x: x.id):
            handle_tree_structure(base_xml_container, root_node)
        return xml_object_dict

    def handle_data_section(xml_qa_export: etree._Element, xml_checkrun_first: etree._Element,
                            xml_checkrun_obj: dict[etree._Element, classes.Object],
                            xml_checkrun_last: etree._Element) -> None:
        def get_name() -> str:
            """Transorms native IFC Attributes like IfcType into desite Attributes"""

            pset_name = obj.ident_attrib.property_set.name
            if pset_name == "IFC":
                return obj.ident_attrib.name

            else:
                return f"{pset_name}:{obj.ident_attrib.name}"

        xml_data_section = etree.SubElement(xml_qa_export, "dataSection")

        check_run_data = etree.SubElement(xml_data_section, "checkRunData")
        check_run_data.set("refID", str(xml_checkrun_first.attrib.get("ID")))
        etree.SubElement(check_run_data, "checkSet")

        for xml_checkrun, obj in xml_checkrun_obj.items():
            check_run_data = etree.SubElement(xml_data_section, "checkRunData")
            check_run_data.set("refID", str(xml_checkrun.attrib.get("ID")))
            filter_list = etree.SubElement(check_run_data, "filterList")
            xml_filter = etree.SubElement(filter_list, "filter")

            xml_filter.set("name", get_name())
            xml_filter.set("dt", "xs:string")
            pattern = f'"{obj.ident_attrib.value[0]}"'  # ToDO: ändern
            xml_filter.set("pattern", pattern)

        check_run_data = etree.SubElement(xml_data_section, "checkRunData")
        check_run_data.set("refID", str(xml_checkrun_last.attrib.get("ID")))
        filter_list = etree.SubElement(check_run_data, "filterList")
        xml_filter = etree.SubElement(filter_list, "filter")
        xml_filter.set("name", "Check_State")
        xml_filter.set("dt", "xs:string")
        xml_filter.set("pattern", '"Ungeprüft"')

    def handle_property_section(xml_qa_export: etree._Element) -> None:
        repository = etree.SubElement(xml_qa_export, "repository")
        property_type_section = etree.SubElement(repository, "propertyTypeSection")
        ptype = etree.SubElement(property_type_section, "ptype")

        ptype.set("key", "1")
        ptype.set("name", "Bestandsdaten:Objekttyp")
        ptype.set("datatype", "xs:string")
        ptype.set("unit", "")
        ptype.set("inh", "true")

        property_section = etree.SubElement(repository, "propertySection")

    def export(path: str) -> None:

        template = handle_template()
        xml_container, xml_qa_export = init_xml(project)
        xml_checkrun_first, xml_attribute_rule_list = define_xml_elements(xml_container, "initial_tests")
        handle_js_rules(xml_attribute_rule_list, "start")
        xml_checkrun_obj = handle_object_rules(xml_container, template)
        xml_checkrun_last, xml_attribute_rule_list = define_xml_elements(xml_container, "untested")
        handle_js_rules(xml_attribute_rule_list, "end")
        handle_data_section(xml_qa_export, xml_checkrun_first, xml_checkrun_obj, xml_checkrun_last)
        handle_property_section(xml_qa_export)

        tree = etree.ElementTree(xml_qa_export)
        with open(path, "wb") as f:
            tree.write(f, xml_declaration=True, pretty_print=True, encoding="utf-8", method="xml")

    if project_tree is None:
        project_tree = project.tree()
    export(path)


def export_bs(project: classes.Project, path: str) -> None:
    def handle_elementsection(xml_parent: etree._Element):

        def handle_section(aggregation: classes.Aggregation, xml_item: etree._Element) -> None:

            nonlocal id_dict
            xml_child = etree.SubElement(xml_item, "section")
            id_dict[aggregation] = aggregation.uuid
            xml_child.set("ID", aggregation.uuid)
            xml_child.set("name", aggregation.object.name)
            xml_child.set("pre", "")
            xml_child.set("type", "typeBsGroup")
            xml_child.set("takt", "")

            for child in sorted(aggregation.children, key=lambda x: x.name):
                connection_type = aggregation.connection_dict[child]
                if connection_type == constants.AGGREGATION:
                    handle_section(child, xml_child)
                else:
                    handle_section(child, xml_item)

        xml_elementsection = etree.SubElement(xml_parent, "elementSection")
        xml_root = etree.SubElement(xml_elementsection, "section")
        xml_root.set("ID", str(uuid.uuid4()))
        xml_root.set("name", "BS Autogenerated")
        xml_root.set("pre", "")
        xml_root.set("type", "typeBsContainer")
        xml_root.set("takt", "")

        root_objects: list[classes.Aggregation] = [aggreg for aggreg in classes.Aggregation if
                                                   aggreg.is_root]

        root_objects.sort(key=lambda x: x.name)

        id_dict = dict()
        for aggreg in root_objects:
            handle_section(aggreg, xml_root)

        return xml_elementsection, id_dict

    def handle_repository(xml_parent: etree._Element, id_dict: dict[classes.Aggregation, str]) -> None:
        def handle_property_type_section() -> dict[str, int]:
            xml_property_type_section = etree.SubElement(xml_repo, "propertyTypeSection")

            attribute_dict = dict()

            i = 1
            for attribute in classes.Attribute:
                # use attribute_text instead of attribute to remove duplicates
                attribute_text = f"{attribute.property_set.name}:{attribute.name}"
                if attribute_text not in attribute_dict:
                    xml_ptype = etree.SubElement(xml_property_type_section, "ptype")
                    xml_ptype.set("key", str(i))
                    xml_ptype.set("name", attribute_text)
                    xml_ptype.set("datatype", attribute.data_type)
                    xml_ptype.set("unit", "")
                    xml_ptype.set("inh", "false")
                    attribute_dict[attribute_text] = i
                    i += 1

            return attribute_dict

        def handle_property_section() -> None:
            xml_property_section = etree.SubElement(xml_repo, "propertySection")

            for node, ref_id in id_dict.items():
                obj = node.object
                for property_set in obj.property_sets:
                    for attribute in property_set.attributes:
                        attribute_text = f"{attribute.property_set.name}:{attribute.name}"
                        ref_type = attribute_dict[attribute_text]
                        xml_property = etree.SubElement(xml_property_section, "property")
                        xml_property.set("refID", str(ref_id))
                        xml_property.set("refType", str(ref_type))
                        if attribute == obj.ident_attrib:
                            xml_property.text = attribute.value[0]
                        else:
                            xml_property.text = "füllen!"

        xml_repo = etree.SubElement(xml_parent, "repository")
        xml_id_mapping = etree.SubElement(xml_repo, "IDMapping")

        for i, (item, id_value) in enumerate(id_dict.items()):
            xml_id = etree.SubElement(xml_id_mapping, "ID")
            xml_id.set("k", str(i + 1))
            xml_id.set("v", str(id_value))

        attribute_dict = handle_property_type_section()
        handle_property_section()

    def handle_relation_section(xml_parent: etree._Element) -> None:
        xml_relation_section = etree.SubElement(xml_parent, "relationSection")

        xml_id_mapping = etree.SubElement(xml_relation_section, "IDMapping")
        xml_relation = etree.SubElement(xml_relation_section, "relation")
        xml_relation.set("name", "default")

    def export() -> None:
        xml_boq_export = handle_header(project, "bsExport")
        xml_elementsection, id_dict = handle_elementsection(xml_boq_export)

        xml_link_section = etree.SubElement(xml_boq_export, "linkSection")
        xml_repository = handle_repository(xml_boq_export, id_dict)
        handle_relation_section(xml_boq_export)

        tree = etree.ElementTree(xml_boq_export)

        with open(path, "wb") as f:
            tree.write(f, xml_declaration=True, pretty_print=True, encoding="utf-8", method="xml")

    if path:
        export()
        pass


def export_bookmarks(proj:classes.Project,path: str) -> None:
    def handle_bookmark_list() -> etree.ElementTree:
        xml_bookmarks = etree.Element("bookmarks")
        xml_bookmarks.set("xmlnsxsi", "http://www.w3.org/2001/XMLSchema-instance")
        xml_bookmark_list = etree.SubElement(xml_bookmarks, "cBookmarkList")

        obj: classes.Object
        for obj in sorted(proj.objects, key=lambda x: x.ident_attrib.value[0]):
            xml_bookmark = etree.SubElement(xml_bookmark_list, "cBookmark")
            xml_bookmark.set("ID", str(obj.uuid))

            if isinstance(obj.ident_attrib, classes.Attribute):
                xml_bookmark.set("name", str(obj.ident_attrib.value[0]))

            xml_bookmark.set("bkmType", "2")
            xml_col = etree.SubElement(xml_bookmark, "col")
            xml_col.set("v", "Type##xs:string")

            attribute = obj.ident_attrib
            xml_col = etree.SubElement(xml_bookmark, "col")
            text = f"{attribute.property_set.name}:{attribute.name}##{attribute.data_type}"
            xml_col.set("v", text)

            for property_set in obj.property_sets:
                for attribute in property_set.attributes:
                    if attribute != obj.ident_attrib:
                        xml_col = etree.SubElement(xml_bookmark, "col")
                        text = f"{property_set.name}:{attribute.name}##{attribute.data_type}"
                        xml_col.set("v", text)
        return etree.ElementTree(xml_bookmarks)

    def get_attribute_dict() -> dict[str,str]:
        attribute_dict = {}
        for obj in proj.objects:
            for property_set in obj.property_sets:
                for attribute in property_set.attributes:
                    attribute_dict[f"{property_set.name}:{attribute.name}"] = attribute.data_type

        return attribute_dict
    if not os.path.isdir(path):
        return


    with open(os.path.join(path, "bookmarks.bkxml"), "wb") as f:
        tree = handle_bookmark_list()
        tree.write(f, xml_declaration=True, pretty_print=True, encoding="utf-8", method="xml")

    attrib_dict = get_attribute_dict()
    file_loader = jinja2.FileSystemLoader(HOME_DIR)
    env = jinja2.Environment(loader=file_loader)
    env.trim_blocks = True
    env.lstrip_blocks = True
    template = env.get_template(BOOKMARK_TEMPLATE)
    code = template.render(attribute_dict=attrib_dict)
    with open(os.path.join(path, "bookmark_script.js"), "w") as f:
        f.write(code)


def export_boq(path: str, pset_name):
    def get_distinct_attributes(property_sets: list[classes.PropertySet]):
        attribute_names = list()

        for property_set in property_sets:
            attribute: classes.Attribute
            attribute_names += [attribute.name for attribute in property_set.attributes]

        distinct_attribute_names = list(dict.fromkeys(attribute_names))

        return distinct_attribute_names

    if path:
        with open(path, "w", ) as file:
            writer = csv.writer(file, delimiter=";")
            property_sets = [property_set for property_set in classes.PropertySet if
                             property_set.name == pset_name]
            distinct_attribute_names = get_distinct_attributes(property_sets)
            header = ["Ident", "Object"] + [f"{pset_name}:{name}" for name in distinct_attribute_names]
            writer.writerow(header)
            obj: classes.Object

            objects = [obj for obj in classes.Object if
                       pset_name in [pset.name for pset in obj.property_sets]]  # find objects with matching propertyset
            for obj in objects:
                property_set = obj.get_property_set_by_name(pset_name)
                ident = obj.ident_attrib
                line = [f"{ident.property_set.name}:{ident.name}", ident.value[0]]
                for attribute_name in distinct_attribute_names:
                    attribute: classes.Attribute = property_set.get_attribute_by_name(attribute_name)

                    if attribute is not None:
                        line.append("|".join(attribute.value))
                    else:
                        line.append("")
                writer.writerow(line)

    pass


def export_attribute_json(project: classes.Project, path):
    with open(path, "w") as file:
        json_dict = dict()
        for obj in sorted(project.objects, key=lambda x: x.ident_value):
            if not obj.property_sets:
                continue
            if obj.ident_value is None:
                continue
            json_dict[obj.ident_value] = dict()
            obj_dict = json_dict[obj.ident_value]
            for property_set in obj.property_sets:
                if not property_set.attributes:
                    continue
                obj_dict[property_set.name] = dict()
                pset_dict = obj_dict[property_set.name]
                for attribute in property_set.attributes:
                    pset_dict[attribute.name] = dict()
                    attribute_dict = pset_dict[attribute.name]

                    attribute_dict[constants.DATA_TYPE] = attribute.data_type
                    if not attribute.value:
                        attribute_dict[constants.VALUE_TYPE] = "Exists"
                    else:
                        attribute_dict[constants.VALUE_TYPE] = attribute.value_type
                    attribute_dict[constants.VALUE] = attribute.value
        json.dump(json_dict, file, indent=1)



from . import constants, classes

def check_attribute_names(project:classes.Project):
    attribute_dict = {} #{lower_case:{attributeName:[attributes],attributename:[attributes]}}
    for obj in project.objects:
        for property_set in obj.property_sets:
            for attribute in property_set.attributes:
                name = attribute.name
                lower_case_name = name.lower()
                if lower_case_name not in attribute_dict:
                    attribute_dict[lower_case_name] = {}

                if name not in attribute_dict[lower_case_name]:
                       attribute_dict[lower_case_name][name] = list()

                attribute_dict[lower_case_name][name].append(attribute)

    problem_attributes = [name_dict for lower,name_dict in attribute_dict.items() if len(name_dict)>1]

    for d in problem_attributes:
        for name,attribute in d.items():
            print(name)
        print("----"*10)


from openpyxl import Workbook
from openpyxl.worksheet.worksheet import Worksheet
from .. import classes, constants


def create_mapping(project: classes.Project, path: str, allplan_mapping_name: str):
    TITLES = ["Definition", "Zuweisung", "Mapping"]

    def transform_datatype(data_type: str) -> str:
        if data_type == constants.XS_INT:
            return "Ganzzahl"
        if data_type == constants.XS_DOUBLE:
            return "Fließkommazahl"

        return "Text"

    def create_definition(worksheet: Worksheet) -> dict[str, classes.Attribute]:
        COLUMNS = ["AttributeName",
                   "AttributeTyp",
                   "AttributeValue",
                   "AttributMin",
                   "AttributMax",
                   "AttrEinh",
                   "AttrEingab",
                   "AttVorgabe_I",
                   "AttVorgabe_II",
                   "AttVorgabe_III",
                   "AttVorgabe_IV"]

        for x, text in enumerate(COLUMNS):
            worksheet.cell(row=1, column=x + 1, value=text)

        attribute_dict: dict[str, classes.Attribute] = dict()

        for attribute in classes.Attribute:
            name = attribute.name
            new_data_type = attribute.data_type
            old_data_type = new_data_type
            old_attribute = attribute_dict.get(name)
            if old_attribute is not None:
                old_data_type = attribute.data_type
            if old_data_type != new_data_type:
                print(f"Achtung bei {attribute.name} neuer Datentyp: {new_data_type}  alter Datentyp: {old_data_type}")
            elif attribute.name not in attribute_dict:
                attribute_dict[attribute.name] = attribute

        for y, attribute in enumerate(attribute_dict.values()):
            row = 2 + y
            worksheet.cell(row=row, column=1, value=attribute.name)
            worksheet.cell(row=row, column=2, value=transform_datatype(attribute.data_type))
            if attribute.data_type == constants.XS_BOOL:
                worksheet.cell(row=row, column=7, value="CheckBox")
        return attribute_dict

    def create_zuweisung(kenner: str, worksheet: Worksheet):

        def get_attrib_count(obj: classes.Object):
            return sum(len([attrib for attrib in pset.attributes]) for pset in obj.property_sets)

        max_attribs = max(get_attrib_count(obj) for obj in project.objects)
        header = ["Kenner"] + ["Wert", "Name"] * max_attribs
        for i, text in enumerate(header):
            worksheet.cell(1, i + 1, text)
        worksheet.cell(2, 1, kenner)

        row_index = 2
        for obj in project.objects:
            if obj.is_concept:
                continue
            worksheet.cell(row_index, 2, obj.ident_value)
            col_index = 3
            for propery_set in obj.property_sets:
                for attribute in propery_set.attributes:
                    if attribute.name != kenner:
                        worksheet.cell(row_index, col_index, attribute.name)
                        col_index += 2
            row_index += 1

    def create_mapping(attribute_dict: dict[str, classes.Attribute], worksheet: Worksheet):
        def transform_type(t: str) -> str:
            if t == constants.XS_INT:
                return "IfcInteger"
            if t == constants.XS_DOUBLE:
                return "IfcReal"
            if t == constants.XS_BOOL:
                return "IfcBoolean"
            return "IfcLabel"

        COLUMNS = ["Objekt", "AttributAllplan", "AttributIfc", "Pset", "Type"]
        for x, text in enumerate(COLUMNS):
            worksheet.cell(row=1, column=x + 1, value=text)
        worksheet.cell(2, 1, "All")
        for index, (name, attribute) in enumerate(sorted(attribute_dict.items())):
            worksheet.cell(2 + index, 2, name)
            worksheet.cell(2 + index, 4, allplan_mapping_name)
            worksheet.cell(2 + index, 5, transform_type(attribute.data_type))

    wb = Workbook()
    ws = wb.active
    ws.title = TITLES[0]

    ad = create_definition(ws)
    create_zuweisung("bauteilKlassifikation", wb.create_sheet(TITLES[1]))
    create_mapping(ad, wb.create_sheet(TITLES[2]))
    wb.save(path)



from __future__ import annotations

import json
import logging
import os.path
import random
import re
import shutil
import tempfile
import uuid
from typing import Iterator

import openpyxl
from openpyxl import Workbook
from openpyxl.cell.cell import Cell
from openpyxl.worksheet.worksheet import Worksheet
from openpyxl.worksheet.table import Table,TableStyleInfo
from openpyxl import styles
from openpyxl.utils import get_column_letter

from .. import classes, constants
from ..Template import IFC_4_1

IDENT_PSET_NAME = "Allgemeine Eigenschaften"
IDENT_ATTRIB_NAME = "bauteilKlassifikation"


def split_string(text: str) -> list[str] | None:
    if text is None:
        return []
    text = re.findall(r'[^,;]+', text)
    for i, item in enumerate(text):
        if "(" in item:
            item = item.split("(")
            text[i] = item[0]
        text[i] = text[i].strip()

    return text


def find_by_abbreviation(abbreviation: str) -> ExcelBlock | None:
    excel_block_dict: dict[str, ExcelBlock] = {block.abbreviation.upper(): block for block in ExcelBlock}
    return excel_block_dict.get(abbreviation.upper())


def autoadjust_column_widths(sheet:Worksheet) -> None:
    for i in range(len(list(sheet.columns))):
        column_letter = get_column_letter(i+1)
        column = sheet[column_letter]
        width = max([len(cell.value) for cell in column if cell.value is not None], default=2)
        sheet.column_dimensions[column_letter].width = width

class ExcelIterator(type):
    def __iter__(cls: ExcelBlock) -> Iterator[ExcelBlock]:
        return iter(cls._registry)


class ExcelBlock(metaclass=ExcelIterator):
    _registry: list[ExcelBlock] = list()

    def __init__(self, base_cell: Cell, sheet: Worksheet):
        self._registry.append(self)
        self.base_cell = base_cell
        self.sheet = sheet
        self.pset: classes.PropertySet | None = None
        self.object: classes.Object | None = None

    @property
    def ident_value(self) -> str | None:
        cell = self.sheet.cell(row=self.base_cell.row,
                               column=self.base_cell.column + 2)
        return cell.value

    @property
    def name(self) -> str:
        cell = self.sheet.cell(row=self.base_cell.row,
                               column=self.base_cell.column + 1)
        return cell.value

    @property
    def abbreviation(self) -> str:
        cell = self.sheet.cell(row=self.base_cell.row + 1,
                               column=self.base_cell.column + 1)
        return cell.value.upper()

    @property
    def is_predefined_pset(self) -> bool:
        if self.ident_value is None:
            return True
        else:
            return False

    @property
    def entry(self) -> Cell:
        if self.is_predefined_pset:
            return self.sheet.cell(row=self.base_cell.row + 4, column=self.base_cell.column)
        else:
            return self.sheet.cell(row=self.base_cell.row + 5, column=self.base_cell.column)

    @property
    def aggregates(self) -> set[ExcelBlock] | None:
        if self.is_predefined_pset:
            return None

        child_cell = self.sheet.cell(row=self.base_cell.row + 3, column=self.base_cell.column + 1)
        child_string_list = split_string(child_cell.value)
        children = set()
        for abbrev in child_string_list:
            if abbrev != "-":
                block = find_by_abbreviation(abbrev)
                if block is not None:
                    children.add(block)
                else:
                    logging.error(f"[{self.name}] abbreviation '{abbrev}' doesn't exist!")
        return children

    @property
    def parent_classes(self) -> set[ExcelBlock] | None:
        parents = set()

        parent_text = self.sheet.cell(self.base_cell.row + 2, self.base_cell.column + 1).value
        for abbrev in split_string(parent_text):
            if abbrev != "-":
                parent_block = find_by_abbreviation(abbrev)
                if parent_block is not None:
                    if parent_block.name != IDENT_PSET_NAME:
                        parents.add(parent_block)
                        parents = set.union(parent_block.parent_classes, parents)
                else:

                    logging.warning(
                        f"[{self.name}] Elternklasse: Kürzel {abbrev.upper()} existiert nicht!")
        return parents

    @property
    def inherits(self) -> set[ExcelBlock]:

        def is_child(value) -> bool:
            last_num = value.split(".")[-1]

            if not last_num.isdigit():
                return False
            last_num = int(last_num)
            if last_num >= 100 and last_num % 100 == 0:
                return True
            else:
                return False

        inherit_blocks = set()

        if self.is_predefined_pset:
            return inherit_blocks

        for block in ExcelBlock:
            if not block.is_predefined_pset:
                parent_txt = block.ident_value.split(".")[:-1]
                parent_txt = ".".join(parent_txt)
                if parent_txt == self.ident_value and is_child(block.ident_value):
                    inherit_blocks.add(block)

        return inherit_blocks

    def create_attributes(self) -> set[classes.Attribute]:
        """
           Iterate over Attributes
           Create Them and find special Datatypes
           """

        def transform_value_types(attribute_name:str,value: str) -> (str, bool):
            if value is not None:
                if value.lower() in ["string", "str"]:
                    data_type = constants.XS_STRING
                elif value.lower() in ["double"]:
                    data_type = constants.XS_DOUBLE
                elif value.lower() in ["boolean", "bool"]:
                    data_type = constants.XS_BOOL
                elif value.lower() in ["int", "integer"]:
                    data_type = constants.XS_INT
                else:
                    text = f"{self.name}: Datatype '{value}' of Attribute {attribute_name} can't be interpreted. Use {constants.XS_STRING} instead "
                    logging.info(text)
                    data_type = constants.XS_STRING
            else:
                data_type = constants.XS_STRING

            return data_type

        entry = self.entry

        cell_list = set(block.base_cell for block in ExcelBlock)
        attributes: set[classes.Attribute] = set()

        while entry.value is not None and entry not in cell_list and entry.value != "-":
            data_type_text = self.sheet.cell(row=entry.row, column=entry.column + 2).value
            data_type = transform_value_types(entry.value,data_type_text)
            attribute_name:str = entry.value
            description = self.sheet.cell(row=entry.row, column=entry.column + 1).value
            optional = False
            if attribute_name.startswith("*"):
                optional = True
                attribute_name = attribute_name[1:]

            attribute = classes.Attribute(self.pset, attribute_name, [], constants.VALUE_TYPE_LOOKUP[constants.LIST],
                                          data_type=data_type,optional=optional,description=description)
            attribute.revit_name =attribute_name
            attributes.add(attribute)

            entry = self.sheet.cell(row=entry.row + 1, column=entry.column)

        return attributes

    def create_predefined_pset(self) -> None:
        optional = False
        pset_name = self.name
        if self.name.startswith("*"):
            optional = True
            pset_name = pset_name[1:]
        self.pset = classes.PropertySet(pset_name,optional=optional)
        self.pset.attributes = self.create_attributes()

    def create_object(self) -> classes.Object:
        pset_name = self.name
        if self.name.startswith("*"):
            pset_name = pset_name[1:]

        self.pset = classes.PropertySet(pset_name)
        attributes = self.create_attributes()
        if not attributes:
            self.pset.delete()
            self.pset = None

        predef_psets: dict[str, ExcelBlock] = {block.name: block for block in ExcelBlock if block.is_predefined_pset}
        parent_pset = predef_psets.get(IDENT_PSET_NAME)

        ident_pset = parent_pset.pset.create_child(IDENT_PSET_NAME)
        ident_attrib = ident_pset.get_attribute_by_name(IDENT_ATTRIB_NAME)
        ident_attrib.value = [self.ident_value]
        optional = False
        obj_name = self.name
        if self.name.startswith("*"):
            optional = True
            obj_name= obj_name[1:]
        obj = classes.Object(obj_name, ident_attrib,ifc_mapping=self.ifc_mapping(),optional=optional)
        if attributes:
            obj.add_property_set(self.pset)
        obj.add_property_set(ident_pset)
        obj.abbreviation = self.abbreviation

        return obj

    def ifc_mapping(self) -> set[str]:
        if self.is_predefined_pset:
            return set()

        cell = self.sheet.cell(row=self.base_cell.row + 4, column=self.base_cell.column + 2)
        value: str = cell.value
        if value is None:
            logging.error(f"[{self.name}]: no IFC Mapping")
            return {"IfcBuildingElementProxy"}

        string_list = value.split("/")
        string_list = [string.strip() for string in string_list]

        for string in string_list:
            if string not in IFC_4_1:
                logging.info(f"[{self.name}]: '{string}' not in IFC 4.1 Specification")
        return set(string_list)


def _create_blocks(sheet) -> None:
    """create Excel Blocks"""
    excel_blocks = set()

    row: tuple[Cell]
    for row in sheet:
        for cell in row:
            if cell.value is not None:
                text = cell.value.strip()
                if text in ["name", "name:"]:
                    if sheet.cell(cell.row + 1, cell.column).value == "Kürzel":
                        excel_blocks.add(ExcelBlock(cell, sheet))
                    else:
                        logging.error(f"{sheet.cell(cell.row + 1, cell.column)} hat den Wert 'name'")


def _create_items() -> None:
    """create Objects and PropertySets"""
    predef_psets = [block for block in ExcelBlock if block.is_predefined_pset]
    objects = [block for block in ExcelBlock if not block.is_predefined_pset]

    for predef_pset_block in predef_psets:
        predef_pset_block.create_predefined_pset()

    for object_block in objects:
        obj = object_block.create_object()
        object_block.object = obj

    for block in objects:
        for pset in [block.pset for block in block.parent_classes if block.pset is not None]:
            new_pset = pset.create_child(pset.name)
            block.object.add_property_set(new_pset)

    #for SOM of Deutsche Bahn
    for obj in classes.Object:
        prop = obj.get_property_set_by_name("Allgemeine Eigenschaften")
        if prop is None:
            continue
        bn = prop.get_attribute_by_name("bauteilName")
        if bn is not None:
            bn.value = [obj.name]

def _build_object_tree() -> None:
    tree_dict: dict[str, classes.Object] = {obj.ident_attrib.value[0]: obj for obj in classes.Object}

    for ident, item in tree_dict.items():
        ident_list = ident.split(".")[:-1]
        parent_obj = tree_dict.get(".".join(ident_list))

        if parent_obj is not None:
            parent_obj.add_child(item)


def _build_aggregations() -> None:
    def get_root_blocks() -> set[ExcelBlock]:
        children = set()
        for e_block in ExcelBlock:  # find all child blocks
            if not e_block.is_predefined_pset:
                children = set.union(e_block.aggregates, children)
                children = set.union(e_block.inherits, children)

        r_blocks = set()  # root blocks
        for e_block in ExcelBlock:
            if e_block not in children and not e_block.is_predefined_pset:
                r_blocks.add(e_block)

        return r_blocks

    def link_child_nodes(aggregation: classes.Aggregation, block: ExcelBlock) -> None:
        aggregate_list = block.aggregates
        inherit_list = block.inherits

        for aggregate_block in aggregate_list:
            if aggregate_block.name != block.name:
                if aggregate_block.is_predefined_pset:
                    logging.error(f"[{block.name}] can't aggregate to {aggregate_block.name}")
                else:
                    child_aggreg = classes.Aggregation(aggregate_block.object)
                    relationship = constants.AGGREGATION
                    if aggregate_block in inherit_list:
                        relationship += constants.INHERITANCE

                    aggregation.add_child(child_aggreg, relationship)
                    link_child_nodes(child_aggreg, aggregate_block)
            else:
                logging.warning(f"[{aggregation.name}] recursive aggregation")

        for inherit_block in inherit_list:
            if inherit_block not in aggregate_list:
                child_aggreg = classes.Aggregation(inherit_block.object)
                aggregation.add_child(child_aggreg, constants.INHERITANCE)
                link_child_nodes(child_aggreg, inherit_block)

    root_blocks = get_root_blocks()

    for block in root_blocks:
        aggregation = classes.Aggregation(block.object)
        link_child_nodes(aggregation, block)


def open_file(path: str, ws_name: str) -> None:
    # TODO: add request for Identification Attribute
    with tempfile.TemporaryDirectory() as tmpdirname:
        new_path = os.path.join(tmpdirname, "excel.xlsx")
        shutil.copy2(path, new_path)
        book = openpyxl.load_workbook(new_path)
        if ws_name not in book.sheetnames:
            logging.error("Worksheet Name not in Workbook")
            return
        sheet = book[ws_name]

        _create_blocks(sheet)
        _create_items()
        _build_object_tree()
        _build_aggregations()

    ExcelBlock._registry = list()


def create_abbreviation_json(excel_path: str, ws_name: str, export_path: str = None) -> dict[str:[str, str]]:
    ExcelBlock._registry = list()
    with tempfile.TemporaryDirectory() as tmpdirname:
        new_path = os.path.join(tmpdirname, "excel.xlsx")
        shutil.copy2(excel_path, new_path)
        book = openpyxl.load_workbook(new_path)
        if ws_name not in book.sheetnames:
            logging.error("Worksheet Name not in Workbook")
            return
        sheet = book[ws_name]

        _create_blocks(sheet)
        _create_items()
        _build_object_tree()
        _build_aggregations()

        d = {block.abbreviation: [block.ident_value, block.name] for block in ExcelBlock if
             block.ident_value is not None}
        if export_path is not None:
            with open(export_path, "w") as file:
                json.dump(d, file, indent=2)
        return d


def export(project: classes.Project, path: str, mapping_dict: dict[str, str] = {}) -> None:

    if not os.path.exists(os.path.dirname(path)):
        raise FileNotFoundError(f"path {os.path.dirname(path)} DNE")

    NAME = "name"
    OBJECTS = "objects"
    TABLE_STYLE = "TableStyleLight1"
    OPTIONAL_FONT = styles.Font(color="4e6ec0")

    def fill_main_sheet(sheet: Worksheet) -> None:
        sheet.title = "Uebersicht"
        sheet.cell(1, 1).value = "bauteilName"
        sheet.cell(1, 2).value = "bauteilKlassifikation"
        row = 1
        for row, obj in enumerate(sorted(project.objects), start=2):
            sheet.cell(row, 1).value = obj.name
            sheet.cell(row, 2).value = str(obj.ident_value)
            if obj.optional:
                sheet.cell(row,1).font = OPTIONAL_FONT
                sheet.cell(row,2).font = OPTIONAL_FONT
        table_start = sheet.cell(1,1).coordinate
        table_end = sheet.cell(row,2).coordinate
        table_range = f"{table_start}:{table_end}"
        table = Table(displayName="Uebersicht",ref=table_range )
        style = TableStyleInfo(name=TABLE_STYLE, showFirstColumn=False,
                               showLastColumn=False, showRowStripes=True, showColumnStripes=False)
        table.tableStyleInfo = style
        sheet.add_table(table)
        autoadjust_column_widths(sheet)

    def filter_to_sheets() -> dict:
        d = {obj.ident_value: {NAME:obj.name,OBJECTS:[]} for obj in project.objects if len(obj.ident_value.split(".")) == 1}
        for ident,name in mapping_dict.items():
            d[ident] = {NAME:name,OBJECTS:[]}

        for obj in project.objects:
            group = obj.ident_value.split(".")[0]
            d[group][OBJECTS].append(obj)

        d["son"]= {NAME:"Sonstige",OBJECTS:[]}
        for group_name,group in list(d.items()):
            objects = group[OBJECTS]
            if len(objects)<3:
                d["son"][OBJECTS]+=objects
                del d[group_name]
        return d

    def create_object_entry(obj:classes.Object, sheet, start_row, start_column,table_index):
        if obj.optional:
            font_style = OPTIONAL_FONT
        else:
            font_style = styles.Font()

        sheet.cell(start_row,start_column).value = "bauteilName"
        sheet.cell(start_row,start_column+1).value = obj.name

        sheet.cell(start_row+1, start_column).value = "bauteilKlassifikation"
        sheet.cell(start_row+1,start_column+1).value = obj.ident_value

        sheet.cell(start_row + 2, start_column).value = "Kürzel"
        sheet.cell(start_row+2,start_column+1).value = str(obj.abbreviation)

        sheet.cell(start_row+3,start_column).value = "Property"
        sheet.cell(start_row+3,start_column+1).value = "Propertyset"
        sheet.cell(start_row+3,start_column+2).value = "Beispiele / Beschreibung"

        for i in range(0,4):
            for k in range(0,3):
                sheet.cell(start_row+i,start_column+k).font = font_style
        draw_border(sheet, [start_row, start_row+2], [start_column, start_column+2])
        fill_grey(sheet, [start_row, start_row+2], [start_column, start_column+2])

        pset_start_row = start_row+4
        index = 0
        for property_set in sorted(obj.property_sets):
            for attribute in sorted(property_set.attributes):
                sheet.cell(pset_start_row+index,start_column).value = attribute.name
                sheet.cell(pset_start_row+index,start_column+1).value = property_set.name
                sheet.cell(pset_start_row+index,start_column+2).value = attribute.description
                if attribute.optional:
                    sheet.cell(pset_start_row + index, start_column).font = OPTIONAL_FONT
                    sheet.cell(pset_start_row + index, start_column+1).font = OPTIONAL_FONT
                    sheet.cell(pset_start_row + index, start_column+2).font = OPTIONAL_FONT
                index+=1

        table_start = sheet.cell(pset_start_row-1,start_column).coordinate
        table_end = sheet.cell(pset_start_row+index-1,start_column+2).coordinate
        table_range = f"{table_start}:{table_end}"
        table = Table(displayName=f"Tabelle_{str(table_index).zfill(5)}",ref=table_range )
        style = TableStyleInfo(name=TABLE_STYLE, showFirstColumn=False,
                               showLastColumn=False, showRowStripes=True, showColumnStripes=False)
        table.tableStyleInfo = style
        sheet.add_table(table)

    def draw_border(sheet:Worksheet,row_range:[int,int],column_range:[int,int]):
        for row in range(row_range[0],row_range[1]+1):
            for column in range(column_range[0],column_range[1]+1):
                left_side = styles.Side(border_style="none",color = "FF000000")
                right_side = styles.Side(border_style="none", color="FF000000")
                top_side = styles.Side(border_style="none", color="FF000000")
                bottom_side = styles.Side(border_style="none", color="FF000000")
                if column == column_range[0]:
                    left_side = styles.Side(border_style="thick", color="FF000000")

                if column == column_range[1]:
                    right_side = styles.Side(border_style="thick", color="FF000000")

                if row == row_range[0]:
                    top_side = styles.Side(border_style="thick", color="FF000000")
                if row == row_range[1]:
                    bottom_side = styles.Side(border_style="thick", color="FF000000")
                sheet.cell(row,column).border = styles.Border(left = left_side,right = right_side,top = top_side,bottom = bottom_side)

    def fill_grey(sheet:Worksheet,row_range:[int,int],column_range:[int,int]):
        fill = styles.PatternFill(fill_type="solid", start_color="d9d9d9")
        for row in range(row_range[0], row_range[1] + 1):
            for column in range(column_range[0], column_range[1] + 1):
                sheet.cell(row, column).fill = fill

    workbook = Workbook()

    sheet_main = workbook.active
    fill_main_sheet(sheet_main)

    sheet_dict = filter_to_sheets()
    table_counter = 1
    for ident,d in sheet_dict.items():
        obj_name = d[NAME]
        objects = d[OBJECTS]
        work_sheet = workbook.create_sheet(f"{obj_name} ({ident})")
        for counter,obj in enumerate(sorted(objects)):
            column = 1+counter*4
            create_object_entry(obj,work_sheet,1,column,table_counter)
            table_counter+=1
        autoadjust_column_widths(work_sheet)
    workbook.save(path)
